""" Handles special commands starting with : to set system parameters or model parameters """
from justai.tools.display import color_print, ERROR_COLOR, SYSTEM_COLOR, ASSISTANT_COLOR
from justai.agent.agent import Agent


class CommandHandler:
    def __init__(self, gpt: Agent):
        self.gpt = gpt

        self.commands = {}
        self.add_command('quit', self.handle_quit, ":quit - Quit the program")
        self.add_command('load', self.handle_load, ":load name - loads the saved conversation with the specified name")
        self.add_command('save', self.handle_save, ":save name - saves the conversation under the specified name")
        self.add_command('debug', self.handle_debug, ":debug - set to True displays all prompts and model replies")
        self.add_command('input', self.handle_input, ":input filename - loads an input from the specified file")
        self.add_command('model', self.handle_gpt_attribute, ":model gpt-4 - Sets the AI model")
        self.add_command('system', self.handle_system, ":system system_message - Sets the system message for the model.")
        self.add_command('max_tokens', self.handle_gpt_attribute,
                         ":max_tokens 800 - The maximum number of tokens to generate in the completion")
        self.add_command('temperature', self.handle_gpt_attribute, ":temperature 0.9 - What sampling temperature to " +
                                                                   "use, between 0 and 2")
        self.add_command('n', self.handle_gpt_attribute, ":n 1 - Specifies the number answers given")
        self.add_command('stop', self.handle_gpt_attribute, ':stop ["\\n", " Human:", " AI:"] - Up to 4 sequences ' +
                                                            'where the API will stop generating further tokens')
        self.add_command('bye', self.handle_bye, ":bye - quits but saves the conversation first")
        self.add_command('help', self.handle_help, ":help - shows this help message")
        for attr in dir(self.gpt):
            if attr[0] != '_' and attr not in self.commands:
                self.add_command(attr, self.handle_gpt_attribute, "")

    def add_command(self, command, handler, description):
        self.commands[command] = (handler, description)

    def handle_quit(self):
        return False

    def handle_load(self, filename):
        if filename:
            self.gpt.load(filename)
        else:
            color_print(f"Pass a filename\n", color=ERROR_COLOR)
        return True

    def handle_input(self, filename):
        if filename:
            self.gpt.file_input(filename)
        else:
            color_print(f"Pass a filename\n", color=ERROR_COLOR)
        return True

    def handle_save(self, filename=None):
        if not filename:
            filename = self.gpt.name
        self.gpt.save(filename)
        color_print(f"Saved to {(self.gpt.save_dir / filename).with_suffix('.txt')}\n", color=SYSTEM_COLOR)
        return True

    def handle_system(self, system_message):
        if not system_message:
            color_print(f"Pass a system message\n", color=ERROR_COLOR)
            return True
        self.gpt.system_message = system_message
        color_print(f"System message set to ", color=SYSTEM_COLOR)
        color_print(f"{system_message}\n", color=ASSISTANT_COLOR)
        return True

    def handle_reset(self):
        self.gpt.reset()
        color_print(f"Conversation reset\n", color=SYSTEM_COLOR)
        return True

    def handle_debug(self, debug: str = None):
        if debug is None:
            color_print(f'debug is {self.gpt.debug}', color=SYSTEM_COLOR)
        else:
            self.gpt.debug = debug.lower() in ['true', '1', 't', 'y', 'yes']
            color_print(f'debug set to {self.gpt.debug}', color=SYSTEM_COLOR)
        return True

    def handle_bye(self):
        self.gpt.save()
        self.gpt.chat('bye')
        return False

    def handle_maxmessages(self, param):
        self.gpt.message_memory = int(param)

    def handle_gpt_attribute(self, attr, value):
        if not value:
            value = getattr(self.gpt, attr)
            color_print(f"{attr} is currently {value}\n", color=SYSTEM_COLOR)
            return True
        try:
            value = eval(value)
        except (SyntaxError, NameError):
            pass  # Treat param as a string
        setattr(self.gpt, attr, value)
        color_print(f"{attr} set to {value}\n", color=SYSTEM_COLOR)
        return True

    def handle_help(self):
        for (_, description) in self.commands.values():
            if description:
                color_print(description, color=SYSTEM_COLOR)
        return True

    def handle_command(self, command: str):
        # Commands can be things like
        # :load filename
        # :save [filename]
        # :quit
        # :max_tokens=1000
        # :temperature=0.8
        if command.count('='):
            command, param = command.split('=', 1)
        elif command.count(' '):
            command, param = command.split(' ', 1)
        else:
            param = None
        if command not in self.commands:
            color_print(f"Unknown command: {command}\n", color=ERROR_COLOR)
            return True
        handler, _ = self.commands[command]
        if handler == self.handle_gpt_attribute:
            return handler(command, param)
        if param:
            return handler(param)
        return handler()



import json


class Message:
    """ Handles the completion as returned by GPT """
    def __init__(self, role=None, text_or_json=None, function_name=None, function_content=None):
        self.role = role
        if not text_or_json:
            self.text = None
            self.data = None
        elif isinstance(text_or_json, str):
            self.text = text_or_json
            self.data = None
        else:
            self.text = json.dumps(text_or_json)
            self.data = text_or_json

    @classmethod
    def from_dict(cls, data: dict):
        message = cls()
        for key, value in data.items():
            setattr(message, key, value)
        return message

    def __bool__(self):
        return bool(self.text)

    def __str__(self):
        res = f'role: {self.role}'
        res += f' text: {self.text}'
        return res

    def to_dict(self):
        dictionary = {}
        for key, value in self.__dict__.items():
            if value is not None:
                dictionary[key] = value
        return dictionary



from pathlib import Path

from dotenv import load_dotenv

from justai.agent.agent import Agent
from justai.tools.prompts import add_prompt_file, get_prompt


class Translator(Agent):

    def __init__(self, model):
        super().__init__(model, temperature=0.3)
        add_prompt_file(Path(__file__).parent / 'prompts.toml')
        self.system_message = get_prompt('SYSTEM')

        self.header = ''
        self.units = []
        self.footer = ''


    def load(self, input_file: str|Path):
        # Input bestaat uit <transunit> elementen. Die hebben een datatype property.
        # Binnen elke <transunit> zit een <source> element en komt (na vertaling) een <target> element.
        # ALs datatype == "plaintext" dan zit de te vertalen tekst direct in de <source>
        # Als datatype == "x-DocumentState" dan zit er in de <source> een <g> element met daarin de te vertalen tekst.
        with open(input_file, 'r') as f:
            return self.read(f.read())

    def read(self, xml: str):
        if not 'urn:oasis:names:tc:xliff:document:1.2' in xml[:300]:
            raise ValueError('Not an XLIFF 1.2 file')
        transunit_blocks = xml.split('<trans-unit')
        self.header = transunit_blocks[0]
        transunit_blocks = transunit_blocks[1:]
        transunit_blocks[-1], self.footer = transunit_blocks[-1].split('</trans-unit>')
        transunit_blocks[-1] += '</trans-unit>'
        self.units = []
        for i in range(len(transunit_blocks)):
            tub = transunit_blocks[i]
            id = tub.split('id="')[1].split('"')[0]
            xml_space = tub.split('xml:space="')[1].split('"')[0] if 'xml:space="' in tub else None
            datatype = tub.split('datatype="')[1].split('"')[0]
            source = tub.split('<source>')[1].split('</source>')[0]
            match datatype:
                case "x-DocumentState":
                    texts = []
                    for block in source.split('<g ')[1:]:
                        t = block.split('">')[1].split('</g>')[0]
                        if _translatable({'text': t}):
                            texts += [t]
                    text = '||'.join(texts)
                    #text = '||'.join([block.split('">')[1].split('</g>')[0] for block in source.split('<g ')[1:]])
                case "plaintext":
                    text = source if _translatable({'text': source}) else ""
                case _:
                    raise ValueError(f'Unknown datatype {datatype}')

            self.units += [{'id': id,
                       'xml_space': xml_space,
                       'datatype': datatype,
                       'source': source,
                       'text': text}] # text is "" als er niks te vertalen is

    def write(self):
        xml = self.header
        for unit in self.units:
            xml_space = f'xml:space="{unit["xml_space"]}" ' if unit.get('xml_space') else ''
            xml += f'''<trans-unit id="{unit['id']}" {xml_space}datatype="{unit['datatype']}">\n'''
            xml += f'''				<source>{unit['source']}</source>\n'''
            xml += '                <target>'
            if unit.get('target'):
                xml += unit['target']
            else:
                xml += unit['source']
            xml += '</target>\n'
            xml += "</trans-unit>\n"
        xml += self.footer
        return xml

    def save_xlf(self, output_file: str | Path):
        with open(output_file, 'w') as f:
            f.write(self.write())

    def translate(self, language: str):
        translatable_units = [u for u in self.units if u['text']]
        translate_list = [f"{index+1} [[{unit['text']}]]\n" for index, unit in enumerate(translatable_units)]
        # TODO: translate_str stap voor stap opbouwen tot ie over de 3000 tekens gaat.
        #  Dan prompt, toevoegen aan translations en repeat
        translate_str = "".join(translate_list)
        prompt = get_prompt('TRANSLATE', language=language, translate_str=translate_str, count=len(translate_list))
        prompt_result = self.chat(prompt, return_json=False)
        translations = [t.split(']]')[0] for t in prompt_result.split('[[')[1:]]
        index = 0
        for unit in self.units:
            if unit['text']:
                translation = translations[index]
                unit['translation'] = translation
                index += 1
                text = unit['text']
                print(text,'->',translation, '\n')
                unit['target'] = unit['source']
                for s, t in zip(text.split('||'), translation.split('||'), strict=True):
                    unit['target'] = unit['target'].replace('>' + s + '<', '>' + t + '<')


def _translatable(unit: dict) -> bool:
    """ Returns True if the unit should be translated """
    return unit['text'] and len(unit['text']) > 1 and unit['text'][0] != '%'


if __name__ == "__main__":
    load_dotenv()
    tr = Translator('gpt-4-turbo-preview')
    tr.load('/Users/hp/ai/opdrachten/Autoniveau/autoniveau/AI.xlf')
    tr.translate('Engels')
    tr.save_xlf('/Users/hp/Downloads/AI_en.xlf')



""" Handles the GPT API and the conversation state. """
import json
import os
import re
from pathlib import Path

import openai
import tiktoken
from openai import APIConnectionError, APIError, RateLimitError, OpenAI
from openai._types import NOT_GIVEN
from dotenv import load_dotenv

from justai.tools.display import print_message, color_print, SYSTEM_COLOR, ERROR_COLOR, DEBUG_COLOR2, DEBUG_COLOR1
from justai.agent.message import Message

BASE_SYSTEM = "You are ChatGPT, a large language model trained by OpenAI."


class Agent:
    def __init__(self, model: str, **kwargs):
        # Authentication
        if not os.getenv("OPENAI_API_KEY"):
            load_dotenv()  # Load the .env file into the environment
        openai.api_key = os.getenv("OPENAI_API_KEY")
        if not openai.api_key:
            color_print("No OpenAI API key found. Create one at https://platform.openai.com/account/api-keys and " +
                        "set it in the .env file like OPENAI_API_KEY=here_comes_your_key.", color=ERROR_COLOR)
        self.client = OpenAI()
        self.functions = {}  # Callable GPT functions
        self.return_type = None  # Structure of the data of the function called by the model. Default is text.

        self.system_message = BASE_SYSTEM

        # Model parameters
        self.model = model

        # The maximum number of tokens to generate in the completion.
        # Defaults to 16
        # The token count of your prompt plus max_tokens cannot exceed the model's context length.
        # Most models have a context length of 2048 tokens (except for the newest models, which support 4096).
        self.max_tokens = kwargs.get('max_tokens', 800)

        # What sampling temperature to use, between 0 and 2.
        # Higher values like 0.8 will make the output more random, while lower values like 0.2
        # will make it more focused and deterministic.
        # We generally recommend altering this or top_p but not both
        # Defaults to 1
        self.temperature = kwargs.get('temperature', 0.5)

        # An alternative to sampling with temperature, called nucleus sampling,
        # where the model considers the results of the tokens with top_p probability mass.
        # So 0.1 means only the tokens comprising the top 10% probability mass are considered.
        # We generally recommend altering this or temperature but not both.
        # Defaults to 1
        self.top_p = kwargs.get('top_p', 1)

        # How many completions to generate for each prompt.
        # Because this parameter generates many completions, it can quickly consume your token quota.
        # Use carefully and ensure that you have reasonable settings for max_tokens and stop.
        self.n = kwargs.get('n', 1)

        # Up to 4 sequences where the API will stop generating further tokens.
        # The returned text will not contain the stop sequence.
        # Example: [" Human:", " AI:"]
        self.stop = kwargs.get('stop')

        # Number between -2.0 and 2.0.
        # Positive values penalize new tokens based on whether they appear in the text so far,
        # increasing the model's likelihood to talk about new topics.
        # Defaults to 0
        self.presence_penalty = kwargs.get('presence_penalty', 0)

        # Number between -2.0 and 2.0.
        # Positive values penalize new tokens based on their existing frequency in the text so far,
        # decreasing the model's likelihood to repeat the same line verbatim.
        # Defaults to 0
        self.frequency_penalty = kwargs.get('frequency_penalty', 0)

        # Parameters to save the current conversation
        self.name = ''  # Name of the current conversation
        self.save_dir = Path(__file__).resolve().parent / 'saves'
        self.save_dir.mkdir(exist_ok=True)
        self.message_memory = 20  # Number of messages to remember. Limits token usage.
        self.messages = []

        self._last_token_count = 0
        self.debug = False

    @classmethod
    def from_json(cls, s, *args, **kwargs):
        agent = cls(None, *args, **kwargs)
        dictionary = json.loads(s)
        for key, value in dictionary.items():
            match key:
                case 'save_dir':
                    agent.save_dir = Path(value)
                case 'messages':
                    agent.messages = [Message.from_dict(m) for m in value]
                case _:
                    setattr(agent, key, value)
        return agent

    def set_api_key(self, key: str):
        """ Used when using Aigent from a browser where the user has to specify a key """
        openai.api_key = key

    def system(self):  # This function can be overwritten by child classes to make the system message dynamic
        return self.system_message

    def reset(self):
        self.name = ''
        self.messages = []

    def get_messages(self):
        result = [{'role': 'system', 'content': self.system()}]
        for m in self.messages[-self.message_memory:]:
            message = {'role': m.role, 'content': m.text}
            result.append(message)
        return result

    def last_token_count(self):
        return self._last_token_count

    def set_return_type(self, model):
        self.return_type = [{
                    "name": model.__name__,
                    "description": model.__doc__,
                    "parameters": model.json_schema()}]

    def chat(self, prompt, add_to_messages=True, return_json=False):

        def chat_completion_request(function_call="auto"):
            if self.debug:
                color_print(f"\nRunning completion with these messages", color=DEBUG_COLOR1)
                for message in self.messages:
                    if hasattr(message, 'text'):
                        color_print(f"{message}", color=DEBUG_COLOR1)
                if self.functions:
                    color_print(f"And these functions", color=DEBUG_COLOR1)
                    for function in self.functions.values():
                        color_print(f"{function.function_name}({function.parameters})", color=DEBUG_COLOR1)
                print()

            messages = self.get_messages()
            last_error = None
            for _ in range(3):
                try:
                    completion = self.client.chat.completions.create(
                        model=self.model,
                        messages=messages,
                        temperature=self.temperature,
                        max_tokens=self.max_tokens,
                        n=self.n,
                        top_p=self.top_p,
                        frequency_penalty=self.frequency_penalty,
                        presence_penalty=self.presence_penalty,
                        stop=self.stop,
                        response_format={"type": "json_object"} if return_json else NOT_GIVEN,
                    )
                    if self.debug and hasattr(completion.choices[0], 'text'):
                        color_print(f"{completion.choices[0].text}", color=DEBUG_COLOR2)
                    self._last_token_count = completion.usage.total_tokens
                    return completion
                except APIConnectionError as e:
                    color_print("Connection error.", color=SYSTEM_COLOR)
                    last_error = e
                except APIError as e:
                    color_print("API error", color=SYSTEM_COLOR)
                    last_error = e
                except RateLimitError as e:
                    color_print(f"{self.model} is overloaded", color=SYSTEM_COLOR)
                    last_error = e
            print('Too many errors. Aborting.')
            raise last_error

        if self.messages and not self.name:
            self.name = re.sub(r'\W+', '', self.messages[0].text).replace(' ', '_')[:20]
        self.messages += [Message('user', prompt)]

        completion = chat_completion_request()
        message_text = completion.choices[0].message.content
        if message_text.startswith('```json'):
            message_text = message_text[7:-3]
            message = Message('assistant', json.loads(message_text))
        else:
            message = Message('assistant', message_text)
        if add_to_messages:
            self.messages += [message]

        return json.loads(message.text) if return_json else message.text

    def after_response(self):
        # content is in messages[-1]['completion']['choices'][0]['message']['content']
        return  # Can be overridden

    def save(self, name=None):
        if name:
            self.name = name

        assert self.name
        with open((self.save_dir / self.name).with_suffix('.txt'), "w") as f:
            f.write(f"system: {self.system()}\n")
            for message in self.messages:
                f.write(f"{message.role}: {message.text}\n")

    def load(self, name):
        def save_message(msg):
            if msg.role == 'system':
                self.system_message = msg.text
            else:
                self.messages += [msg]

        self.messages = []
        self.name = name
        if not name.endswith('.txt'):
            name += '.txt'
        filename = self.save_dir / name
        if not os.path.isfile(filename):
            color_print(f"New conversation:  {filename}", color=SYSTEM_COLOR)
            return
        with open(filename, "r") as f:
            message = Message()
            assert not message
            for line in f.readlines():
                line = line[:-1]
                try:
                    role, text = line.split(': ', 1)
                except ValueError:
                    message.text += '\n' + line
                    continue
                if role in ['system', 'user', 'assistant', 'function']:
                    if message:
                        save_message(message)
                    message = Message(role=role, text_or_json=text)
                else:
                    message.text += '\n' + line
            if message:
                save_message(message)
        print_message(Message('system', self.system()), 'system')
        for message in self.messages:
            print_message(message.text, message.role)

    def file_input(self, filename):
        with open(filename, "r") as f:
            prompt = f.read()
        self.chat(prompt)

    def dumps(self) -> str:
        data = {}
        for key, value in self.__dict__.items():
            if value is not None:
                try:
                    json.dumps(value)
                    data[key] = value
                except (TypeError, ValueError):
                    match key:
                        case 'save_dir':
                            data[key] = str(value)
                        case 'messages':
                            data[key] = [message.to_dict() for message in value]
        return json.dumps(data, indent=2)

    def token_count(self, text: str):
        encoding = tiktoken.encoding_for_model(self.model)
        return len(encoding.encode(text))
